@@ -1,5 +1,5 @@
 /*
-  Copyright 2007-2012 David Robillard <http://drobilla.net>
+  Copyright 2007-2015 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -22,10 +22,12 @@
 #include "./suil_config.h"
 #include "./suil_internal.h"
 
-#define GTK2_UI_URI LV2_UI__GtkUI
-#define QT4_UI_URI  LV2_UI__Qt4UI
-#define X11_UI_URI  LV2_UI__X11UI
-#define WIN_UI_URI  LV2_UI_PREFIX "WindowsUI"
+#define GTK2_UI_URI  LV2_UI__GtkUI
+#define QT4_UI_URI   LV2_UI__Qt4UI
+#define QT5_UI_URI   LV2_UI__Qt5UI
+#define X11_UI_URI   LV2_UI__X11UI
+#define WIN_UI_URI   LV2_UI_PREFIX "WindowsUI"
+#define COCOA_UI_URI LV2_UI__CocoaUI
 
 SUIL_API
 unsigned
@@ -43,11 +45,17 @@ suil_ui_supported(const char* container_type_uri,
 	            && !strcmp(ui_type_uri, QT4_UI_URI))
 	           || (!strcmp(container_type_uri, QT4_UI_URI)
 	               && !strcmp(ui_type_uri, GTK2_UI_URI))
+	           || (!strcmp(container_type_uri, QT5_UI_URI)
+	               && !strcmp(ui_type_uri, GTK2_UI_URI))
 	           || (!strcmp(container_type_uri, GTK2_UI_URI)
 	               && !strcmp(ui_type_uri, X11_UI_URI))
 	           || (!strcmp(container_type_uri, GTK2_UI_URI)
 	               && !strcmp(ui_type_uri, WIN_UI_URI))
+	           || (!strcmp(container_type_uri, GTK2_UI_URI)
+	               && !strcmp(ui_type_uri, COCOA_UI_URI))
 	           || (!strcmp(container_type_uri, QT4_UI_URI)
+	               && !strcmp(ui_type_uri, X11_UI_URI))
+	           || (!strcmp(container_type_uri, QT5_UI_URI)
 	               && !strcmp(ui_type_uri, X11_UI_URI))) {
 		return SUIL_WRAPPING_EMBEDDED;
 	} else {
@@ -62,27 +70,55 @@ open_wrapper(SuilHost*      host,
              LV2_Feature*** features,
              unsigned       n_features)
 {
-	if (!strcmp(container_type_uri, ui_type_uri)) {
-		return NULL;
-	}
-
 	const char* module_name = NULL;
+#ifdef SUIL_WITH_GTK2_IN_QT4
 	if (!strcmp(container_type_uri, QT4_UI_URI)
 	    && !strcmp(ui_type_uri, GTK2_UI_URI)) {
 		module_name = "suil_gtk2_in_qt4";
-	} else if (!strcmp(container_type_uri, GTK2_UI_URI)
-	           && !strcmp(ui_type_uri, QT4_UI_URI)) {
+	}
+#endif
+#ifdef SUIL_WITH_GTK2_IN_QT5
+	if (!strcmp(container_type_uri, QT5_UI_URI)
+	    && !strcmp(ui_type_uri, GTK2_UI_URI)) {
+		module_name = "suil_gtk2_in_qt5";
+	}
+#endif
+#ifdef SUIL_WITH_QT4_IN_GTK2
+	if (!strcmp(container_type_uri, GTK2_UI_URI)
+	    && !strcmp(ui_type_uri, QT4_UI_URI)) {
 		module_name = "suil_qt4_in_gtk2";
-	} else if (!strcmp(container_type_uri, GTK2_UI_URI)
-	           && !strcmp(ui_type_uri, X11_UI_URI)) {
+	}
+#endif
+#ifdef SUIL_WITH_X11_IN_GTK2
+	if (!strcmp(container_type_uri, GTK2_UI_URI)
+	    && !strcmp(ui_type_uri, X11_UI_URI)) {
 		module_name = "suil_x11_in_gtk2";
-	} else if (!strcmp(container_type_uri, GTK2_UI_URI)
-	           && !strcmp(ui_type_uri, WIN_UI_URI)) {
+	}
+#endif
+#ifdef SUIL_WITH_WIN_IN_GTK2
+	if (!strcmp(container_type_uri, GTK2_UI_URI)
+	    && !strcmp(ui_type_uri, WIN_UI_URI)) {
 		module_name = "suil_win_in_gtk2";
-	} else if (!strcmp(container_type_uri, QT4_UI_URI)
-	           && !strcmp(ui_type_uri, X11_UI_URI)) {
+	}
+#endif
+#ifdef SUIL_WITH_COCOA_IN_GTK2
+	if (!strcmp(container_type_uri, GTK2_UI_URI)
+	    && !strcmp(ui_type_uri, COCOA_UI_URI)) {
+		module_name = "suil_cocoa_in_gtk2";
+	}
+#endif
+#ifdef SUIL_WITH_X11_IN_QT4
+	if (!strcmp(container_type_uri, QT4_UI_URI)
+	    && !strcmp(ui_type_uri, X11_UI_URI)) {
 		module_name = "suil_x11_in_qt4";
 	}
+#endif
+#ifdef SUIL_WITH_X11_IN_QT5
+	if (!strcmp(container_type_uri, QT5_UI_URI)
+	    && !strcmp(ui_type_uri, X11_UI_URI)) {
+		module_name = "suil_x11_in_qt5";
+	}
+#endif
 
 	if (!module_name) {
 		SUIL_ERRORF("Unable to wrap UI type <%s> as type <%s>\n",
@@ -108,6 +144,7 @@ open_wrapper(SuilHost*      host,
 	void* lib = dlopen(path, RTLD_NOW);
 	if (!lib) {
 		SUIL_ERRORF("Unable to open wrap module %s (%s)\n", path, dlerror());
+		free(path);
 		return NULL;
 	}
 
@@ -197,7 +234,7 @@ suil_instance_new(SuilHost*                 host,
 	instance->descriptor = descriptor;
 
 	// Make UI features array
-	instance->features = (LV2_Feature**)malloc(sizeof(LV2_Feature**));
+	instance->features = (LV2_Feature**)malloc(sizeof(LV2_Feature*));
 	instance->features[0] = NULL;
 
 	// Copy user provided features
@@ -230,8 +267,15 @@ suil_instance_new(SuilHost*                 host,
 	}
 
 	// Open wrapper (this may add additional features)
-	instance->wrapper = open_wrapper(
-		host, container_type_uri, ui_type_uri, &instance->features, n_features);
+	if (container_type_uri && strcmp(container_type_uri, ui_type_uri)) {
+		instance->wrapper = open_wrapper(host,
+		                                 container_type_uri, ui_type_uri,
+		                                 &instance->features, n_features);
+		if (!instance->wrapper) {
+			suil_instance_free(instance);
+			return NULL;
+		}
+	}
 
 	// Instantiate UI
 	instance->handle = descriptor->instantiate(
@@ -285,16 +329,27 @@ suil_instance_free(SuilInstance* instance)
 			instance->descriptor->cleanup(instance->handle);
 		}
 
+		dlclose(instance->lib_handle);
+
 		// Close libraries and free everything
 		if (instance->wrapper) {
+#ifndef _WIN32
+			// Never unload modules on windows, causes mysterious segfaults
 			dlclose(instance->wrapper->lib);
+#endif
 			free(instance->wrapper);
 		}
-		dlclose(instance->lib_handle);
 		free(instance);
 	}
 }
 
+SUIL_API
+SuilHandle
+suil_instance_get_handle(SuilInstance* instance)
+{
+	return instance->handle;
+}
+
 SUIL_API
 LV2UI_Widget
 suil_instance_get_widget(SuilInstance* instance)
